# -*- coding: utf-8 -*-
"""
Created on Fri May  1 01:55:16 2020
@author: yoelr
"""
from . import utils
import numpy as np
from numpy import linalg
from collections import deque

__all__ = ('as_least_squares_iter',
           'LeastSquaresIteration',
           'LstSqIter',
)

@utils.njitable
def weighted_average(xs, weights):
    weights /= weights.sum()
    return xs @ weights

def compute_weighted_average_by_least_squares(A, xs):
    b = 1e-12 * np.ones(A.shape[0])
    weights = linalg.lstsq(A, b, None)[0]
    return weighted_average(xs, weights)

class LeastSquaresIteration:
    __slots__ = ('guess_history',
                 'error_history', 
                 'N_activate',
                 '_counter', )
    
    def __init__(self, N_history=5, N_activate=20):
        self.guess_history = deque(maxlen=N_history)
        self.error_history = deque(maxlen=N_history)
        self.N_activate = N_activate
        self._counter = 0

    def __call__(self, x, fx):
        guess_history = self.guess_history
        error_history = self.error_history
        guess_history.append(x)
        error_history.append(fx - x)
        if self.active:
            A = np.array(error_history, dtype=float)
            A = A.transpose()
            xs = np.array(guess_history, dtype=float).transpose()
            return compute_weighted_average_by_least_squares(A, xs)
    
    def reset(self):
        self.guess_history.clear()
        self.error_history.clear()
        self._counter = 0
    
    @property
    def active(self):
        active = self._counter == self.N_activate
        if not active: self._counter += 1
        return active
    
    
LstSqIter = LeastSquaresIteration

def fake_least_squares_iter(x, fx): return fx

def as_least_squares_iter(lstsq):
    if lstsq: 
        if not isinstance(lstsq, LstSqIter):
            lstsq = LstSqIter()
    else:
        lstsq = fake_least_squares_iter
    return lstsq


# -*- coding: utf-8 -*-
"""
Created on Tue Apr  7 09:25:27 2020

@author: yoelr
"""
from numba import njit
from numba.extending import register_jitable
from . import fast
import sys

__all__ = ('njitable', 'njit_alternative', 'speed_up')

#: All njitable functions.
njitables = []

#: All njitable functions that are not replaced in the origin module.
njit_alternatives = []

def njitable(f=None, **options):
    """
    Decorate function as njitable. All 'njitable' functions must be 
    compilable by Numba's njit decorator.
    
    Notes
    -----
    When `flexsolve.speed_up` is run, all njitable functions are compiled.
    
    """
    if not f: return lambda f: njitable(f, **options)
    f_jitable = register_jitable(**options)(f) if options else register_jitable(f)
    njitables.append((f, options))
    return f_jitable

def njit_alternative(f=None, **options):
    """
    Decorate function as njit-alternative. All 'njit_alternative' functions
    must be compilable by Numba's njit decorator. All njit-alternative 
    functions are saved in the `flexsolve.fast` module.
    
    Notes
    -----
    When `flexsolve.speed_up` is run, all njit-alternative functions in 
    the `flexsolve.fast` module are compiled.
    
    """
    if not f: return lambda f: njitable(f, **options)
    f_jitable = register_jitable(**options)(f) if options else register_jitable(f)
    njit_alternatives.append((f, options))
    setattr(fast, f.__name__, f_jitable)
    return f_jitable

def speed_up():
    """
    Speed up simulations by jit compiling all functions registered as
    'njitable'.
    
    See also
    --------
    njitable
    njit_alternative
    
    """
    setfield = setattr
    for f, options in njitables:
        f_jit =  njit(f, **options)
        setfield(sys.modules[f.__module__], f.__name__, f_jit)
        setfield(fast, f.__name__, f_jit)
    njitables.clear()
    for f, options in njit_alternatives:
        setfield(fast, f.__name__,  njit(f, **options))
    njit_alternatives.clear()


# -*- coding: utf-8 -*-
"""
Created on Sun Apr 26 15:43:30 2020

@author: yoelr
"""


def __getattr__(name):
    from flexsolve import speed_up
    import sys 
    
    speed_up()
    dct = sys.modules[__name__].__dict__
    if name in dct:
        return dct[name]
    else:
        raise AttributeError(f"module {__name__} has no attribute {name}")


# -*- coding: utf-8 -*-
"""
Created on Tue Nov 19 22:47:51 2019

@author: yoelr
"""
from . import jit_speed
from . import open_solvers
from . import bounded_solvers
from . import iterative_solvers
from . import least_squares_iteration
from . import profiler
from . import utils
from . import fast

__all__ = (*open_solvers.__all__,
           *bounded_solvers.__all__,
           *iterative_solvers.__all__,
           *least_squares_iteration.__all__,
           *profiler.__all__,
           *jit_speed.__all__,
           'utils',
           'fast',
)

from .jit_speed import *
from .open_solvers import *
from .bounded_solvers import *
from .iterative_solvers import *
from .least_squares_iteration import *
from .profiler import *


# -*- coding: utf-8 -*-
"""
Created on Tue Nov 19 22:50:26 2019

@author: yoelr
"""
from flexsolve.jit_speed import njit_alternative
from flexsolve import utils

__all__ = ('secant', 'aitken_secant')

@njit_alternative
def secant(f, x0, x1=None, xtol=None, ytol=5e-8, args=(), maxiter=50,
           checkroot=True):
    """Secant solver."""
    if x1 is None: x1 = x0 + 1e-5
    y0 = f(x0, *args)
    abs_ = abs
    if abs_(y0) < ytol: return x0
    dx = x1-x0 
    for iter in range(maxiter): 
        y1 = f(x1, *args)
        if abs_(y1) < ytol: return x1
        if y1 == y0 or (xtol and abs_(dx) < xtol): break 
        x1 = x0 - y1*dx/(y1-y0)
        dx = x1-x0 
        x0 = x1
        y0 = y1
    utils.raise_root_error(checkroot)
    return x1

@njit_alternative
def aitken_secant(f, x0, x1=None, xtol=None, ytol=5e-8, args=(), maxiter=50,
                  checkroot=True):
    """Secant solver with Aitken acceleration."""
    if x1 is None: x1 = x0 + 1e-5
    abs_ = abs
    y0 = f(x0, *args)
    if abs_(y0) < ytol: return x0
    dx = x1-x0
    aitken_iter = utils.scalar_aitken_iter
    for iter in range(maxiter):
        y1 = f(x1, *args)
        if abs_(y1) < ytol: return x1
        if y1 == y0 or (xtol and abs_(dx) < xtol): break 
        x0 = x1 - y1*dx/(y1-y0) # x0 = g
        dx = x0-x1
        y0 = y1
        y1 = f(x0, *args)
        if abs_(y1) < ytol: return x0
        if y1 == y0 or (xtol and abs_(dx) < xtol): x1 = x0; break 
        x2 = x0 - y1*dx/(y1-y0) # x2 = gg
        dx = x1 - x0 # x - g
        x1 = aitken_iter(x1, x2, dx, x2 - x0)
        dx = x1 - x0
        y0 = y1
    utils.raise_root_error(checkroot)
    return x1


# -*- coding: utf-8 -*-
"""
Created on Tue Apr 28 16:56:59 2020

@author: yoelr
"""
import numpy as np
from flexsolve.jit_speed import njit_alternative
from flexsolve.least_squares_iteration import as_least_squares_iter
from . import utils


__all__ = ('fixed_point',
           'conditional_fixed_point',
           'wegstein',
           'conditional_wegstein',
           'aitken',
           'conditional_aitken',
           'fixed_point_lstsq',
) 

def fixed_point_lstsq(f, x, xtol=5e-8, args=(), maxiter=50, lstsq=True,
                      checkroot=True):
    """The least-squares solution of a matrix of prior iterations is partially
    used to iteratively esmitate the root."""
    lstsq = as_least_squares_iter(lstsq)
    x0 = x1 = x
    for iter in range(maxiter):
        if x0 is None:
            x0 = x1
            x1 = f(x0, *args)
        else:
            try: x1 = f(x0, *args)
            except:
                x0 = x1
                x1 = f(x0)
        if (np.abs(x1 - x0) < xtol).all(): return x1
        x0 = lstsq(x0, x1)
    utils.raise_root_error(checkroot)
    return x1

@njit_alternative
def fixed_point(f, x, xtol=5e-8, args=(), maxiter=50, checkroot=True):
    """Iterative fixed-point solver."""
    x0 = x1 = x
    for iter in range(maxiter):
        x1 = f(x0, *args)
        if (np.abs(x1 - x0) < xtol).all(): return x1
        x0 = x1
    utils.raise_root_error(checkroot)
    return x1

@njit_alternative
def conditional_fixed_point(f, x):
    """Conditional iterative fixed-point solver."""
    x0 = x1 = x
    condition = True
    while condition:
        x1, condition = f(x0)
        x0 = x1

@njit_alternative
def wegstein(f, x, xtol=5e-8, args=(), maxiter=50, checkroot=True):
    """Iterative Wegstein solver."""
    x0 = x
    x1 = g0 = f(x0, *args)
    wegstein_iter = utils.wegstein_iter
    for iter in range(maxiter):
        dx = x1 - x0
        try: g1 = f(x1, *args)
        except:
            x1 = g0
            g1 = f(x1, *args)
        if utils.fixedpoint_converged(g1 - x1, xtol): return g1
        x0 = x1
        x1 = wegstein_iter(x1, dx, g1, g0)
        g0 = g1
    utils.raise_root_error(checkroot)
    return g1

@njit_alternative
def conditional_wegstein(f, x):
    """Conditional iterative Wegstein solver."""
    x0 = x
    g0, condition = f(x0)
    g1 = x1 = g0
    wegstein_iter = utils.wegstein_iter
    while condition:
        try: g1, condition = f(x1)
        except:
            x1 = g1
            g1, condition = f(x1)
        g1 = g1
        dx = x1-x0
        x0 = x1
        x1 = wegstein_iter(x1, dx, g1, g0)
        g0 = g1

@njit_alternative
def aitken(f, x, xtol=5e-8, args=(), maxiter=50, checkroot=True):
    """Iterative Aitken solver."""
    gg = x
    aitken_iter = utils.aitken_iter
    for iter in range(maxiter):
        try: g = f(x, *args)
        except:
            x = gg
            g = f(x, *args)
        dxg = x - g
        if utils.fixedpoint_converged(dxg, xtol): return g
        gg = f(g, *args)
        dgg_g = gg - g
        if utils.fixedpoint_converged(dgg_g, xtol): return gg
        x = aitken_iter(x, gg, dxg, dgg_g)
    utils.raise_root_error(checkroot)
    return gg

@njit_alternative
def conditional_aitken(f, x):
    """Conditional iterative Aitken solver."""
    condition = True
    gg = x
    aitken_iter = utils.aitken_iter
    while condition:
        try:
            g, condition = f(x)
        except:
            x = gg.copy()
            g, condition = f(x)
        if not condition: return g
        gg, condition = f(g)
        x = aitken_iter(x, gg, x - g, gg - g)


# -*- coding: utf-8 -*-
"""
Created on Mon Apr 20 16:41:04 2020

@author: yoelr
"""
from flexsolve.jit_speed import njitable
from numba.extending import overload, register_jitable
from collections.abc import Iterable
from numba import types
import numpy as np

__all__ = ('pick_best_solution', 'wegstein_iter',
           'aitken_iter', 'array_wegstein_iter', 'scalar_wegstein_iter',
           'array_aitken_iter', 'scalar_aitken_iter', 'not_within_bounds',
           'iteration_is_getting_stuck', 'bisect', 'false_position_iter',
           'IQ_iter', 'fixedpoint_converged')

np.seterr(divide='raise', invalid='raise')

@njitable(cache=True)
def pick_best_solution(xys):
    y_best = 1e16
    x_best = 0.
    abs_ = abs
    for x, y in xys:
        y = abs_(y)
        if y < y_best:
            y_best = y
            x_best = x
    return x_best

def wegstein_iter(x, dx, g1, g0):
    if isinstance(x, Iterable):
        return array_wegstein_iter(x, dx, g1, g0)
    else:
        return scalar_wegstein_iter(x, dx, g1, g0)

def aitken_iter(x, gg, dxg, dgg_g):
    if isinstance(x, Iterable):
        return array_aitken_iter(x, gg, dxg, dgg_g)
    else:
        return scalar_aitken_iter(x, gg, dxg, dgg_g)

@overload(wegstein_iter)
def jit_wegstein_iter(x, dx, g1, g0):
    if isinstance(x, types.Array) and x.ndim:
        return array_wegstein_iter  
    else:
        return scalar_wegstein_iter

@overload(aitken_iter)
def jit_aitken_iter(x, gg, dxg, dgg_g):
    if isinstance(x, types.Array) and x.ndim:
        return array_aitken_iter
    else:
        return scalar_aitken_iter

@register_jitable(cache=True)
def array_wegstein_iter(x, dx, g1, g0):
    denominator = dx-g1+g0
    mask = np.logical_and(np.abs(denominator) > 1e-16, np.abs(dx) < 1e16)
    w = np.ones_like(dx)
    w[mask] = dx[mask]/denominator[mask]
    return w*g1 + (1.-w)*x

@register_jitable(cache=True)
def scalar_wegstein_iter(x, dx, g1, g0):
    denominator = dx-g1+g0
    if abs(denominator) > 1e-16 and abs(dx) < 1e16:
        w = dx / denominator
        x = w*g1 + (1.-w)*x
    else:
        x = g1
    return x
        
@register_jitable(cache=True)
def scalar_fixedpoint_converged(dx, xtol):
    return abs(dx) < xtol

@register_jitable(cache=True)
def array_fixedpoint_converged(dx, xtol):
    return (np.abs(dx) < xtol).all()

def fixedpoint_converged(dx, xtol):
    if isinstance(dx, Iterable) and dx.ndim:
        return array_fixedpoint_converged(dx, xtol)
    else:
        return scalar_fixedpoint_converged(dx, xtol)

@overload(fixedpoint_converged)
def fixedpoint_converged(dx, xtol):
    if isinstance(dx, types.Array) and dx.ndim:
        return array_fixedpoint_converged
    else:
        return scalar_fixedpoint_converged

@register_jitable(cache=True)
def array_aitken_iter(x, gg, dxg, dgg_g):
    denominator = dgg_g + dxg
    mask = np.logical_and(np.abs(denominator) > 1e-16, np.abs(dxg) < 1e16)
    x = x.copy()
    x[mask] -= dxg[mask]**2./denominator[mask]
    nmask = np.logical_not(mask)
    x[nmask] = gg[nmask]
    return x

@register_jitable(cache=True)
def scalar_aitken_iter(x, gg, dxg, dgg_g):
    denominator = dgg_g + dxg
    if abs(denominator) > 1e-16 and abs(dxg) < 1e16:
        x -= dxg**2. / denominator 
    else:
        x = gg
    return x

@njitable(cache=True)
def not_within_bounds(x, x0, x1):
    return not (x0 < x < x1 or x1 < x < x0)

@njitable(cache=True)
def iteration_is_getting_stuck(x, xlast, dx, r=0.1):
    return abs((x - xlast) / dx) < r

@njitable(cache=True)
def bisect(x0, x1):
    return (x0 + x1) / 2.0

@njitable(cache=True)
def false_position_iter(x0, x1, dx, y0, y1, df, xlast):
    dy = y1 - y0
    if dy:
        x = x0 + df*dx/dy
        if not_within_bounds(x, x0, x1):
            x = bisect(x0, x1)
        elif iteration_is_getting_stuck(x, xlast, dx):
            x = (x + x0 + x1) / 3.
    else:
        x = bisect(x0, x1)
    return x

@njitable(cache=True)
def IQ_iter(y0, y1, y2, x0, x1, x2, dx, df0, xlast):
    df1 = -y1
    df2 = -y2
    d01 = df0-df1
    d02 = df0-df2
    d12 = df1-df2
    if abs(d01) > 1e-32 and abs(d02) > 1e-32 and abs(d12) > 1e-32:
        df0_d12 = df0 / d12
        df1_d02 = df1 / d02
        df2_d01 = df2 / d01
        x = x0*df1_d02*df2_d01 - x1*df0_d12*df2_d01 + x2*df0_d12*df1_d02    
        if not_within_bounds(x, x0, x1):
            x = bisect(x0, x1)
    else:
        x = false_position_iter(x0, x1, dx, y0, y1, df0, xlast)
    return x

@njitable(cache=True)
def raise_root_error(not_satisfied):
    if not_satisfied:
        raise RuntimeError('root could not be solved within error tolerance')
        
@njitable(cache=True)
def raise_bounds_error(y0, y1):
    if y0 * y1 > 0.:
        raise RuntimeError('f(x0) and f(x1) must have opposite signs')


# -*- coding: utf-8 -*-
"""
Created on Wed Nov 20 23:09:53 2019

@author: yoelr
"""
import numpy as np
import matplotlib.pyplot as plt
from copy import copy

__all__ = ('Profiler',)
            
            
class Archive:
    __slots__ = ('name', 'xs', 'ys')
    
    def __init__(self, name, xs, ys):
        self.name = name
        self.xs = np.array(xs, float)
        self.ys = np.array(ys, float)
    
    def __len__(self):
        return len(self.xs)
    
    @property
    def size(self):
        return self.xs.size
    
    @property
    def x_min(self):
        return self.xs.min()
    
    @property
    def x_max(self):
        return self.xs.max()
    
    @property
    def y_min(self):
        return self.ys.min()
    
    @property
    def y_max(self):
        return self.ys.max()
    
    def __repr__(self):
        return f"{type(self).__name__}({repr(self.name)}, xs={self.xs}, ys={self.ys})"

    def _ipython_display_(self):
        print(f"{type(self).__name__}(\n"
              f"    {repr(self.name)},\n"
              f"    xs={self.xs},\n"
              f"    ys={self.ys}\n"
              ")")


class Profiler:
    __slots__ = ('f', 'xs', 'ys', 'archives')
    def __init__(self, f, ):
        self.f = f
        self.archives = []
        self.xs = []
        self.ys = []
        
    def __call__(self, x, *args):
        x = copy(x)
        self.xs.append(x)
        y = self.f(x, *args)
        self.ys.append(y)
        return y
    
    def archive(self, name):
        self.archives.append(Archive(name, self.xs, self.ys))
        self.xs = []
        self.ys = []

    def sizes(self):
        return {archive.name: len(archive) for archive in self.archives}
    
    def _plot_points(self, rxs, rys, offset, step):
        cycle = plt.rcParams['axes.prop_cycle']
        colors = cycle.by_key()['color']
        archives = self.archives
        for color, archive in zip(colors, archives):
            xs = archive.xs
            ys = archive.ys + offset
            plt.scatter(xs, ys, color=color,
                        label=f"{archive.name} ({archive.size} iterations)")
            plt.plot(rxs, rys + offset, color=color, alpha=0.85)
            offset -= step

    def plot(self, title=None, args=(), markbounds=True,
             plot_outside_bounds=True, N=50, shade=True,
             remove_ticks=True):
        plt.figure()
        archives = self.archives
        archives.sort(key=lambda x: x.size)
        x_mins = np.array([i.x_min for i in archives])
        x_maxs = np.array([i.x_max for i in archives])
        x_min = x_mins.min()
        x_max = x_maxs.max()
        dx = (x_max - x_min) / 50 if plot_outside_bounds else 0.
        xs = np.linspace(x_min - dx, x_max + dx, N)
        f = lambda x: self.f(x, *args)
        ys = np.array([f(x) for x in xs])
        y_min = ys.min()
        y_max = ys.max()
        N = len(archives)
        X = max(12 - 2*N, 3)
        offset = (y_max - y_min) / X
        step = 2 * offset / N
        offset -= step / 2
        self._plot_points(xs, ys, offset, step)
        if shade:
            plt.fill_between(xs, ys - offset - step, ys + offset + step,
                             color='grey', alpha=0.1)

        x_solution = np.mean([i.xs[-1] for i in archives])        
        y_solution = f(x_solution)
        y_lb, y_ub = plt.ylim()
        
        x_start = archives[0].xs[0]
        if markbounds and np.all(x_min == x_mins) and np.all(x_max == x_maxs):
            plt.vlines([x_min, x_solution, x_max],
                       [y_lb, y_lb, y_lb],
                       [f(x_min) + offset + step, y_solution + offset + step, f(x_max) + offset + step],
                       linestyles = 'dashed',
                       color='grey')
            plt.xticks([x_min, x_solution, x_max], 
                       [f'{x_min:.3g}\nlower\nbound',
                        f'{x_solution:.3g}\nsolution',
                        f'{x_max:.3g}\nupper\nbound'])
        elif np.all(np.array([i.xs[0] for i in archives]) == x_start):
            plt.vlines([x_start, x_solution],
                       [y_lb, y_lb],
                       [f(x_start) + offset + step,
                        y_solution + offset + step],
                       linestyles = 'dashed',
                       color='grey')
            plt.xticks([x_start, x_solution],
                       [f'{x_start:.3g}\nguess',
                        f'{x_solution:.3g}\nsolution'])
        else:
            plt.vlines([x_solution],
                       [y_lb],
                       [y_solution],
                       linestyles = 'dashed',
                       color='grey')
            plt.xticks([x_solution], ['solution'])
        
        if title: plt.title(title)
        plt.ylim([y_lb, y_ub])
        plt.xlim([xs[0], xs[-1]])
        if remove_ticks:
            plt.tick_params(axis='both', which='both', length=0)
            plt.yticks([])
        plt.legend()
        
    
    def __repr__(self):
        return f"{type(self).__name__}({self.f})"


# -*- coding: utf-8 -*-
"""
Created on Tue Jul  9 00:35:01 2019

@author: yoelr
"""
import numpy as np
from .jit_speed import njit_alternative
from . import utils
__all__ = ('false_position', 'bisection', 
           'IQ_interpolation', 'find_bracket')

# %% Tools

@njit_alternative
def find_bracket(f, x0, x1, y0=-np.inf, y1=np.inf, args=(), maxiter=50):
    """
    Return a bracket within `x0` and `x1` where the objective function, `f`, is 
    certain to have a root.
    """
    bisect = utils.bisect
    if y1 < 0.:  x1, y1, x0, y0 = x0, y0, x1, y1
    isfinite = np.isfinite
    for iter in range(maxiter):
        x = bisect(x0, x1)
        y = f(x, *args)
        if y > 0.:
            x1 = x
            y1 = y
        else:
            x0 = x
            y0 = y
        if isfinite(y0) and isfinite(y1): return (x0, x1, y0, y1)
    raise RuntimeError('failed to find bracket')
    
# %% Solvers

@njit_alternative
def false_position(f, x0, x1, y0=None, y1=None, x=None,
                   xtol=None, ytol=5e-8, args=(), maxiter=50,
                   checkroot=True, checkbounds=True):
    """False position solver."""
    if x is None: x = 1e32
    if y0 is None: y0 = f(x0, *args)
    if y1 is None: y1 = f(x1, *args)
    if y1 < 0.:  x1, y1, x0, y0 = x0, y0, x1, y1
    abs_ = abs
    err0 = abs_(y0)
    err1 = abs_(y1)
    if err0 < err1:
        x_best = x0
        err_best = err0
    else:
        x_best = x1
        err_best = err1
    if checkbounds and y0 * y1 > 0.:
        raise ValueError('f(x0) and f(x1) must have opposite signs')
    dx = x1 - x0
    df = - y0
    false_position_iter = utils.false_position_iter
    if utils.not_within_bounds(x, x0, x1):
        x = false_position_iter(x0, x1, dx, y0, y1, df, x0)
    for iter in range(maxiter):
        y = f(x, *args)
        if y > ytol:
            x1 = x
            err = y1 = y
        elif y < -ytol:
            x0 = x
            y0 = y
            err = df = -y0
        else: return x
        if err < err_best:
            err_best = err
            x_best = x
        dx = x1 - x0
        if xtol and abs_(dx) < xtol: break
        x = false_position_iter(x0, x1, dx, y0, y1, df, x)
    if x_best != x: f(x, *args)
    utils.raise_root_error(checkroot and err_best < ytol)
    return x_best

@njit_alternative
def bisection(f, x0, x1, y0=None, y1=None, x=None, xtol=None, ytol=5e-8, args=(),
              maxiter=50, checkroot=True, checkbounds=True):
    """Bisection solver."""
    if y0 is None: y0 = f(x0, *args)
    if y1 is None: y1 = f(x1, *args)
    if y1 < 0.:  x1, y1, x0, y0 = x0, y0, x1, y1
    dx = x1 - x0
    abs_ = abs
    err0 = abs_(y0)
    err1 = abs_(y1)
    if err0 < err1:
        x_best = x0
        err_best = err0
    else:
        x_best = x1
        err_best = err1
    if checkbounds and y0 * y1 > 0.:
        raise ValueError('f(x0) and f(x1) must have opposite signs')
    bisect = utils.bisect
    if x is None: x = bisect(x0, x1)
    nytol = -ytol
    for iter in range(maxiter):
        y = f(x, *args)
        if y > ytol:
            x1 = x
            err = y
        elif y < nytol:
            x0 = x
            err = -y
        else: return x
        if err < err_best:
            err_best = err
            x_best = x
        dx = x1 - x0
        if xtol and abs_(dx) < xtol: break
        x = bisect(x0, x1)
    if x_best != x: f(x, *args)
    utils.raise_root_error(checkroot and err_best < ytol)
    return x_best

@njit_alternative
def IQ_interpolation(f, x0, x1, y0=None, y1=None, x=None,
                     xtol=None, ytol=5e-8, args=(), maxiter=50,
                     checkroot=True, checkbounds=True):
    """Inverse quadratic interpolation solver."""
    abs_ = abs
    if y0 is None: y0 = f(x0, *args)
    if y1 is None: y1 = f(x1, *args)
    if x is None: x = 1e32
    if y1 < 0.: x1, y1, x0, y0 = x0, y0, x1, y1
    df0 = -y0
    dx = x1 - x0
    if utils.not_within_bounds(x, x0, x1):
        x = utils.false_position_iter(x0, x1, dx, y0, y1, df0, x0)
    err0 = abs_(y0)
    err1 = abs_(y1)
    if err0 < err1:
        x_best = x0
        err_best = err0
    else:
        x_best = x1
        err_best = err1
    if checkbounds and y0 * y1 > 0.:
        raise ValueError('f(x0) and f(x1) must have opposite signs')
    nytol = -ytol
    for iter in range(maxiter):
        y = f(x, *args)
        if y > ytol:
            y2 = y1
            x2 = x1
            x1 = x
            err = y1 = y
        elif y < nytol:
            y2 = y0
            x2 = x0
            x0 = x
            y0 = y
            err = df0 = -y
        else: return x
        if err < err_best:
            err_best = err
            x_best = x
        dx = x1 - x0
        if xtol and abs_(dx) < xtol: break
        x = utils.IQ_iter(y0, y1, y2, x0, x1, x2, dx, df0, x)
    if x_best != x: f(x, *args)
    utils.raise_root_error(checkroot and err_best < ytol)
    return x_best

    
    


# -*- coding: utf-8 -*-
"""
Created on Sat Nov 18 16:17:00 2017

@author: Yoel Cortes-Pena
"""
from setuptools import setup

setup(
    name='flexsolve',
    packages=['flexsolve'],
    license='MIT',
    version='0.3.17',
    description='Flexible function solvers',
    long_description=open('README.rst').read(),
    author='Yoel Cortes-Pena',
    install_requires=['numba>=0.50.0', 'llvmlite>=0.31', 'numpy'],
    python_requires=">=3.6",
    package_data=
        {'flexsolve': []},
    platforms=['Windows', 'Mac', 'Linux'],
    author_email='yoelcortes@gmail.com',
    url='https://github.com/yoelcortes/flexsolve',
    download_url='https://github.com/yoelcortes/flexsolve.git',
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Console',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Scientific/Engineering :: Mathematics'],
    keywords='solve equation function flexible',
)